"""
FMH3 format information for pyfmh3
"""

from construct import Struct, Tell, Const, Padding, Int32ul, RepeatUntil, CString, Pointer, Byte, Int16ul, Flag, Int64ul, Bytes, Seek, If

_construct_version = None
try:
    from construct import __version__ as _construct_version
    _construct_version = [int(v) for v in _construct_version.split('.')]
except Exception:
    # I'd rather just continue than throw an error if this fails for some reason, like versioning changes,
    # so just let _construct_version be None
    # Users following instructions should never have a low version anyway
    pass

if _construct_version:
    if (_construct_version[0] < 2) or ((_construct_version[0] == 2) and (_construct_version[1] < 9)):
        raise Exception('Construct version too low, please install version 2.9+')


_fmh3_format = Struct(
    "pointer_offset" / Tell,
    "signature" / Const(b'FMH3'),
    Padding(4),
    "fonts_count" / Int32ul,
    "fonts_pointers_offset" / Int32ul,
    "fonts" / Pointer(lambda this: this.fonts_pointers_offset + this.pointer_offset, RepeatUntil(lambda obj,lst,ctx: ctx._index >= ctx.fonts_count - 1, Struct(
        "pointer" / Int32ul,
        "data" / Pointer(lambda this: this.pointer + this._.pointer_offset, Struct(
            "id" / Int32ul,
            "advance_width" / Byte,
            "line_height" / Byte,
            "box_width" / Byte,
            "box_height" / Byte,
            "layout_param_1" / Byte, 
            "layout_param_2_numerator" / Byte,
            "layout_param_2_denominator" / Byte,
            Padding(1),
            "other_params?" / Int32ul,
            "tex_size_chars" / Int32ul,
            "chars_count" / Int32ul,
            "chars_pointer" / Int32ul,
            "chars" / Pointer(lambda this: this.chars_pointer + this._._.pointer_offset, RepeatUntil(lambda obj,lst,ctx: ctx._index >= ctx.chars_count - 1, Struct(
                "codepoint" / Int16ul,
                "halfwidth" / Flag,
                Padding(1),
                "tex_col" / Byte,
                "tex_row" / Byte,
                "glyph_x" / Byte,
                "glyph_width" / Byte,
            ))),
        )),
    ))),
)

_fmh3_int64_format = Struct(
    "pointer_offset" / Tell,
    "signature" / Const(b'FMH3'),
    Padding(4),
    "fonts_count" / Int64ul,
    "fonts_pointers_offset" / Int64ul,
    "fonts" / Pointer(lambda this: this.fonts_pointers_offset + this.pointer_offset, RepeatUntil(lambda obj,lst,ctx: ctx._index >= ctx.fonts_count - 1, Struct(
        "pointer" / Int64ul,
        "data" / Pointer(lambda this: this.pointer + this._.pointer_offset, Struct(
            "id" / Int32ul,
            "advance_width" / Byte,
            "line_height" / Byte,
            "box_width" / Byte,
            "box_height" / Byte,
            "layout_param_1" / Byte, 
            "layout_param_2_numerator" / Byte,
            "layout_param_2_denominator" / Byte,
            Padding(1),
            "other_params?" / Int32ul,
            "tex_size_chars" / Int32ul,
            "chars_count" / Int32ul,
            "chars_pointer" / Int32ul,
            "chars" / Pointer(lambda this: this.chars_pointer + this._._.pointer_offset, RepeatUntil(lambda obj,lst,ctx: ctx._index >= ctx.chars_count - 1, Struct(
                "codepoint" / Int16ul,
                "halfwidth" / Flag,
                Padding(1),
                "tex_col" / Byte,
                "tex_row" / Byte,
                "glyph_x" / Byte,
                "glyph_width" / Byte,
            ))),
        )),
    ))),
)

_pof1_struct = Struct(
    "pointer_offset" / Tell,
    "signature" / Const(b'POF1'),
    "section_size" / Int32ul,
    "data_pointer" / Const(32, Int32ul),
    "flags" / Const(b'\x00\x00\x00\x10'),
    "depth" / Const(0, Int32ul),
    "data_size" / Int32ul,
    Padding(8),
    # hardcoded offset for this instead of pointer
    "data" / Bytes(lambda this: this.data_size)
)

_eofc_struct = Struct(
    "pointer_offset" / Tell,
    "signature" / Const(b'EOFC'),
    "section_size" / Const(0, Int32ul),
    "data_pointer" / Const(32, Int32ul),
    "flags" / Const(b'\x00\x00\x00\x10'),
    "depth" / Const(0, Int32ul),
    "data_size" / Const(0, Int32ul),
    Padding(8)
)

_fonm_format = Struct(
    "pointer_offset" / Tell,
    "signature" / Const(b'FONM'),
    "section_size" / Int32ul,
    "data_pointer" / Int32ul,
    "flags" / Const(b'\x00\x00\x00\x10'),
    "depth" / Const(0, Int32ul),
    "data_size" / Int32ul,
    "data" / Pointer(lambda this: this.data_pointer + this.pointer_offset, _fmh3_int64_format),
    # Seek(lambda this: this.data_pointer + this.data_size),
    # ENRS (endian reversal) ignored
    "extra_sections" / If(lambda this: this._building, Struct(
        "pof1" / Pointer(lambda this: this._.data_pointer + this._.data_size + this._.pointer_offset, _pof1_struct),
        Pointer(lambda this: this._.data_pointer + this._.data_size + this._.pointer_offset + this.pof1_size, _eofc_struct),
        Pointer(lambda this: this._.data_pointer + this._.section_size + this._.pointer_offset, _eofc_struct)
    )),
)

_fmh3_types = {
    'FMH3': {
        'remarks': 'unencapsulated FT fontmap',
        'struct': _fmh3_format,
        'address_size': 4,
        'nest_fmh3_data': False,
    },
    'FONM': {
        'remarks': 'X fontmap in FONM container',
        'struct': _fonm_format,
        'address_size': 8,
        'nest_fmh3_data': True,
    },
}


"""
Helper for when you want to load a file that may or may not be inside a farc archive.
"""

from pydiva import pyfarc

def farc_load_helper(s, filenames):
    """
    Takes a stream and list of desired files, returns a list of tuples [(filename, bytes)]
    If stream s does not contain a supported farc file, returns original bytes with filenamee None.
    Otherwise, opens the archive and returns the requested files if present.
    (files not present in the archive will be ignored)
    """
    
    try:
        farc = pyfarc.from_stream(s, files_whitelist=filenames)
    except pyfarc.UnsupportedFarcTypeException:
        return [(None, s.read())]
    
    outlist = []
    for fname, info in farc['files'].items():
        outlist += [(fname, info['data'])]
    
    return outlist


"""
Farc format information for pyfarc
"""

from construct import Struct, Const, Int32ub, Int32sb, RepeatUntil, CString, Pointer, Bytes, Padding, BitStruct, Flag, IfThenElse, Seek, Tell

_construct_version = None
try:
    from construct import __version__ as _construct_version
    _construct_version = [int(v) for v in _construct_version.split('.')]
except Exception:
    # I'd rather just continue than throw an error if this fails for some reason, like versioning changes,
    # so just let _construct_version be None
    # Users following instructions should never have a low version anyway
    pass

if _construct_version:
    if (_construct_version[0] < 2) or ((_construct_version[0] == 2) and (_construct_version[1] < 9)):
        raise Exception('Construct version too low, please install version 2.9+')


_FArc_format = Struct(
    "signature" / Const(b'FArc'),
    "header_size" / Int32ub, # doesn't include signature or header_size
    "alignment" / Int32sb,
    "files" / RepeatUntil(lambda obj,lst,ctx: ctx._io.tell() - 7 > ctx.header_size, Struct(
        "name" / CString("utf8"),
        "pointer" / Int32ub,
        "size" / Int32ub,
        "data" / Pointer(lambda this: this.pointer, Bytes(lambda this: this.size))
    )),
    #Padding(lambda this: this.alignment - (this._io.tell() % this.alignment) if this._io.tell() % this.alignment else 0)
)

_FArC_format = Struct(
    "signature" / Const(b'FArC'),
    "header_size" / Int32ub, # doesn't include signature or header_size
    "alignment" / Int32sb,
    "files" / RepeatUntil(lambda obj,lst,ctx: ctx._io.tell() - 7 > ctx.header_size, Struct(
        "name" / CString("utf8"),
        "pointer" / Int32ub,
        "compressed_size" / Int32ub,
        "uncompressed_size" / Int32ub,
        "data" / Pointer(lambda this: this.pointer, Bytes(lambda this: this.compressed_size))
    )),
    #Padding(lambda this: this.alignment - (this._io.tell() % this.alignment) if this._io.tell() % this.alignment else 0)
)

_FARC_format = Struct(
    "signature" / Const(b'FARC'),
    "header_size" / Int32ub, # doesn't include signature or header_size
    "flags" / BitStruct(
        Padding(29),
        "encrypted" / Flag,
        "compressed" / Flag,
        Padding(1)
    ),
    Padding(4),                     # if not encrypted or else popcnt of alignment is 1, use format field
    "alignment" / Int32sb,          # (if is encrypted and popcnt of alignment is not 1, assume FT format)
    "format" / Const(0, Int32sb),   # this struct only supports DT
    Padding(4),
    "files" / RepeatUntil(lambda obj,lst,ctx: ctx._io.tell() - 7 > ctx.header_size, Struct(
        "name" / CString("utf8"),
        "pointer" / Int32ub,
        "compressed_size" / Int32ub,
        "uncompressed_size" / Int32ub,
        "data" / Pointer(lambda this: this.pointer, Bytes(lambda this: (this.compressed_size + 16 - (this.compressed_size % 16)) if (this.compressed_size % 16 and this._.flags.encrypted) else (this.compressed_size)))
    )),
    #Padding(lambda this: this.alignment - (this._io.tell() % this.alignment) if this._io.tell() % this.alignment else 0)
)

_FARC_FT_format = Struct(
    "signature" / Const(b'FARC'),
    "header_size" / Int32ub, # doesn't include signature or header_size
    "flags" / BitStruct(
        Padding(29),
        "encrypted" / Flag,
        "compressed" / Flag,
        Padding(1)
    ),
    Padding(4),                     # if not encrypted or else popcnt of alignment is 1, use format field
    "alignment" / Int32sb,          # (if is encrypted and popcnt of alignment is not 1, assume FT format)
    "format" / Const(1, Int32sb),   # this struct only supports FT with unencrypted header
    "entry_count" / Int32sb,
    IfThenElse(lambda this: this._parsing, Padding(4), Const(16, Int32sb)),
    "files" / RepeatUntil(lambda obj,lst,ctx: (ctx._io.tell() - 7 > ctx.header_size) or (ctx._index >= ctx.entry_count-1), Struct(
        "name" / CString("utf8"),
        "pointer" / Int32ub,
        "compressed_size" / Int32ub,
        "uncompressed_size" / Int32ub,
        "flags" / BitStruct(
            Padding(29),
            "encrypted" / Flag,
            "compressed" / Flag,
            Padding(1)
        ),
        "io_pos" / Tell, # save stream position
        "file_end" / IfThenElse(lambda this: this._parsing, Seek(lambda this: this._io.seek(0, 2)), Seek(0x7fffffff)), # dirty trick using Seek for end of file when parsing and max int for building
        Seek(lambda this: this.io_pos), # restore position
        "data" / Pointer(lambda this: this.pointer, Bytes(lambda this: (this.compressed_size + 16 - (this.compressed_size % 16)) if (this.compressed_size % 16 and this.flags.encrypted) else (this.compressed_size)))
    )),
    #Padding(lambda this: this.alignment - (this._io.tell() % this.alignment) if this._io.tell() % this.alignment else 0)
)

_farc_types = {
    'FArc': {
        'remarks': 'basic farc format',
        'struct': _FArc_format,
        'compression_support': False,
        'compression_forced': False,
        'fixed_header_size': 4,
        'files_header_fields_size': 8,
        'has_flags': False,
        'has_per_file_flags': False,
        'encryption_type': None,
        'write_support': True,
        'encryption_write_support': False,
        'format_field': None,
    },
    'FArC': {
        'remarks': 'farc with compression support',
        'struct': _FArC_format,
        'compression_support': True,
        'compression_forced': True,
        'fixed_header_size': 4,
        'files_header_fields_size': 12,
        'has_flags': False,
        'has_per_file_flags': False,
        'encryption_type': None,
        'write_support': True,
        'encryption_write_support': False,
        'format_field': None,
    },
    'FARC': {
        'remarks': 'farc with encryption and compression support (DT/F/X)',
        'struct': _FARC_format,
        'compression_support': True,
        'compression_forced': False,
        'fixed_header_size': 20,
        'files_header_fields_size': 12,
        'has_flags': True,
        'has_per_file_flags': False,
        'encryption_type': 'DT',
        'write_support': True,
        'encryption_write_support': True,
        'format_field': ['FARC', 'FARC_FT'],
    },
    'FARC_FT': {    # note: FARC_FT is an internal name only -- reading and writing should use FARC with format 1
        'remarks': 'farc with encryption and compression support (FT)',
        'struct': _FARC_FT_format,
        'compression_support': True,
        'compression_forced': False,
        'fixed_header_size': 24,
        'files_header_fields_size': 16,
        'has_flags': True,
        'has_per_file_flags': True,
        'encryption_type': 'FT',
        'write_support': True,
        'encryption_write_support': True,
        'format_field': ['FARC', 'FARC_FT'],
    },
}